#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include "debug.h"
#include "ptfps.h"
#include "list.h"
#include "get_date.h"
#include "signame.h"

int optcmp(const void *a, const void *b)
{
    const option_t *x = a, *y = b;

    return strcasecmp(x->name, y->name);
}

/* Generic constructors */
static int opt_parse_null(char *args[], void **ret)
{
    return 0;
}

static int opt_parse_intlist(char *args[], void **ret)
{
    int n;
    char *next;
    char *prev = *args;
    list_t *list;

    list = list_create();
    if (list == NULL)
        return -1;

    while (1) {
        next = strchr(prev, ',');
        if (sscanf(prev, "%d", &n) != 1) {
            list_destroy(list);
            return -1;
        }
        if (list_add(list, &n, sizeof(n)) < 0) {
            list_destroy(list);
            return -1;
        }
        if (next == NULL)
            break;
        prev = next + 1;
    }

    *ret = list;
    return 0;
}

static int opt_parse_strlist(char *args[], void **ret)
{
    char *next;
    char *prev = *args;
    list_t *list;

    list = list_create();
    if (list == NULL)
        return -1;

    while (1) {
        next = strchr(prev, ',');
        if (next == NULL) {
            if (list_add(list, prev, strlen(prev)+1) < 0) {
                list_destroy(list);
                return -1;
            }
            break;
        } else {
            *next = '\0';
            if (list_add(list, prev, (next-prev)+1) < 0) {
                list_destroy(list);
                return -1;
            }
        }
        prev = next + 1;
    }

    *ret = list;
    return 0;
}

static int opt_parse_int(char *args[], void **ret)
{
    int *i;

    i = malloc(sizeof(int));
    if (i == NULL)
        return -1;

    *ret = i;
    if (sscanf(*args, "%d", i) != 1) {
        free(i);
        return -1;
    }

    return 0;
}

static int opt_parse_ulong(char *args[], void **ret)
{
    unsigned long *ul;

    ul = malloc(sizeof(unsigned long));
    if (ul == NULL)
        return -1;

    *ret = ul;
    if (sscanf(*args, "%lu", ul) != 1) {
        free(ul);
        return -1;
    }

    return 0;
}

static int opt_parse_double(char *args[], void **ret)
{
    double *d;

    d = malloc(sizeof(double));
    if (d == NULL)
        return -1;

    *ret = d;
    if (sscanf(*args, "%lf", d) != 1) {
        free(d);
        return -1;
    }

    return 0;
}

static int opt_parse_time(char *args[], void **ret)
{
    time_t *t, tmp;

    t = malloc(sizeof(time_t));
    if (t == NULL)
        return -1;

    time(&tmp);
    *t = get_date(*args, &tmp);
    if (*t < (time_t)0) {
        free(t);
        return -1;
    }

    return 0;
}

static int opt_parse_print(char *args[], void **ret)
{
    *ret = *args;

    return 0;
}

/* -true */
static int opt_eval_true(expression_t *e, process_t *p, char *buf, size_t size)
{
    return !(e->negated);
}

/* -false */
static int opt_eval_false(expression_t *e, process_t *p, char *buf, size_t size)
{
    return e->negated;
}

/* -euid, -uid */
static int opt_eval_euid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(uid_t *)(l->data) == get_euid(p))
            return 1;
    }

    return 0;
}

/* -ruid */
static int opt_eval_ruid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(uid_t *)(l->data) == get_ruid(p))
            return 1;
    }

    return 0;
}

/* -suid */
static int opt_eval_suid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(uid_t *)(l->data) == get_suid(p))
            return 1;
    }

    return 0;
}

/* -user */
static int opt_eval_user(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (str_match((char *)l->data, get_eusername(p)))
            return 1;
    }

    return 0;
}

/* -egid, -gid */
static int opt_eval_egid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(gid_t *)(l->data) == get_egid(p))
            return 1;
    }

    return 0;
}

/* -rgid */
static int opt_eval_rgid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(gid_t *)(l->data) == get_rgid(p))
            return 1;
    }

    return 0;
}
/* -sgid */
static int opt_eval_sgid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(gid_t *)(l->data) == get_sgid(p))
            return 1;
    }

    return 0;
}


/* -group */
static int opt_eval_group(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (str_match((char *)l->data, get_rgroupname(p)))
            return 1;
    }

    return 0;
}

/* -name */
static int opt_eval_name(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (str_match((char *)l->data, get_name(p)))
            return 1;
    }

    return 0;
}

/* -tty */
static int opt_eval_tty(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (str_match((char *)l->data, get_name(p)))
            return 1;
    }

    return 0;
}

/* -pid */
static int opt_eval_pid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(pid_t *)(l->data) == get_pid(p))
            return 1;
    }

    return 0;
}

/* -ppid */
static int opt_eval_ppid(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (*(pid_t *)(l->data) == get_ppid(p))
            return 1;
    }

    return 0;
}

/* -stime */
static int opt_eval_stime(expression_t *e, process_t *p, char *buf, size_t size)
{
    return *(time_t *)(e->term_value) < get_stime(p);
}

/* -time */
static int opt_eval_time(expression_t *e, process_t *p, char *buf, size_t size)
{
    return *(unsigned long *)(e->term_value) < get_time(p);
}	

/* -pcpu */
static int opt_eval_pcpu(expression_t *e, process_t *p, char *buf, size_t size)
{
    return *(double *)(e->term_value) < get_pcpu(p);
}	

/* -pmem */
static int opt_eval_pmem(expression_t *e, process_t *p, char *buf, size_t size)
{
    return *(double *)(e->term_value) < get_pmem(p);
}	

/* -print */
static int opt_eval_print(expression_t *e, process_t *p, char *buf, size_t size)
{
    pt_debug("in eval_print%s\n", e->printed_header ? "" : " (printing header)");
    if (e->printed_header) {
        print_process(p, buf, size, e->term_value, 0);
    } else {
        print_process(p, buf, size, e->term_value, 1);
        e->printed_header = 1;
    }

    return 1;
}

/* -kill */
static int opt_parse_kill(char *args[], void **ret)
{
    int i;
    int *sig;
    char *sigstr;

    sig = malloc(sizeof(int));
    if (sig == NULL)
        return -1;
    *ret = sig;

    sigstr = *args;
    if (sscanf(sigstr, "%d", sig) == 1)
        return 0;

    i = signame_to_signum(sigstr);
    if (i < 0) {
        free(sig);
        return -1;
    }

    *sig = i;
    return 0;
}

static int opt_eval_kill(expression_t *e, process_t *p, char *buf, size_t size)
{
    return kill_process(p, *(int *)e->term_value);
}

/* -exec */
static int opt_parse_exec(char *args[], void **ret)
{
    char **tmp;
    list_t *list;

    list = list_create();
    if (list == NULL)
        return -1;

    for (tmp = args; (*tmp != NULL) && (strcmp(*tmp, ";") != 0); tmp++) {
        if (list_add(list, *tmp, strlen(*tmp) + 1) < 0) {
            list_destroy(list);
            return -1;
        }
    }

    *ret = list;
    return 0;
}

static int opt_eval_exec(expression_t *e, process_t *p, char *buf, size_t size)
{
    return exec_process(p, e->term_value);
}

/* -cmd */
static int opt_eval_cmd(expression_t *e, process_t *p, char *buf, size_t size)
{
    list_node_t *l;

    for (l = ((list_t *)(e->term_value))->head; l != NULL; l = l->next){
        if (str_match((char *)l->data, get_cmd(p)))
            return 1;
    }

    return 0;
}

/* -nice */
static int opt_eval_nice(expression_t *e, process_t *p, char *buf, size_t size)
{
    return (get_nice(p) >= *(int *)e->term_value);
}

/* -pri */
static int opt_eval_pri(expression_t *e, process_t *p, char *buf, size_t size)
{
    return (get_pri(p) >= *(int *)e->term_value);
}

/* -runnable */
static int opt_eval_runnable(expression_t *e, process_t *p, char *buf, size_t size)
{
    return get_state(p) == ST_RUNNABLE;
}

/* -sleeping */
static int opt_eval_sleeping(expression_t *e, process_t *p, char *buf, size_t size)
{
    return get_state(p) == ST_SLEEPING;
}
/* -stopped */
static int opt_eval_stopped(expression_t *e, process_t *p, char *buf, size_t size)
{
    return get_state(p) == ST_STOPPED;
}
/* -zombie */
static int opt_eval_zombie(expression_t *e, process_t *p, char *buf, size_t size)
{
    return get_state(p) == ST_ZOMBIE;
}

option_t options[] = {
    {
        "-true", 
        " - Always evaluates to true",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_true
    },
    {"-false", 
        " - Always evaluates to false",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_false
    },
    {
        "-uid",
        " <uid1>[,uid2[,uid3...]] - Evaluates to true when "
            "the effective UID (user id) of a process matches the given uids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_euid
    },
    {
        "-user", 
        " <user1>[,user2[,user3...]] - Evaluates to true when the effective "
            "user name of a process matches any of the given usernames.",
        1,
        OPT_TEST,
        opt_parse_strlist,
        opt_eval_user
    },
    {
        "-gid",
        " <gid1>[,gid2[,gid3...]] - Evaluates to true when the effective GID "
            "(group id) of a process matches any of the given gids",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_egid
    },
    {
        "-group",
        " <group1>[,group2[,group3...]] - Evaluates to true when the effective "
            "group name of a process matches any of the given group names.",
        1,
        OPT_TEST,
        opt_parse_strlist,
        opt_eval_group
    },
    {
        "-name", 
        " <name1>[,name2[,name3]] - Evaluates to true when the name of a "
            "process matches any of the given process names.  In this case, the "
            "name of a process refers to the executable name as well as the "
            "initial few arguments of the process.",
        1,
        OPT_TEST,
        opt_parse_strlist,
        opt_eval_name
    },
    {
        "-tty",
        " <tty1>[,tty2[,tty3]] - Evaluates to true when the tty name of a "
            "process matches any of the given tty names.",
        1,
        OPT_TEST,
        opt_parse_strlist,
        opt_eval_tty
    },
    {
        "-pid",
        " <pid1>[,pid2[,pid3...]] - Evaluates to true when the PID (process "
            "id) of a process matches any of the given pids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_pid
    },
    {
        "-ppid", 
        " <ppid1>[,ppid2[,ppid3...]] - Evaluates to true when the parent PID"
            "(process id) of a process matches any of the given pids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_ppid
    },
    {
        "-stime", 
        " <date> - Evaluates to true on processes whose start time is "
            "later than the given date.",
        1,
        OPT_TEST,
        opt_parse_time,
        opt_eval_stime
    },
    {
        "-time",
        " <seconds> - Evaluates to true on processes that have used more CPU "
            "time than the given number of seconds.",
        1,
        OPT_TEST,
        opt_parse_ulong,
        opt_eval_time
    },
    {
        "-pcpu",
        " <percentage> - Evaluates to true on processes that have used the "
            "CPU (as a percentage of its total running time) more than the "
            "given percentage.",
        1,
        OPT_TEST,
        opt_parse_double,
        opt_eval_pcpu
    },
    {
        "-pmem",
        " <percentage> -  Evaluates to true on processes that are using more "
            "memory (as a percentage of total system memory on the host of the "
            "process) than the given percentage.",
        1,
        OPT_TEST,
        opt_parse_double,
        opt_eval_pmem
    },
    {
        "-print", 
        " <format> - Always evaluates to true and prints the given process "
            "in the given format.  The format can be NORMAL, LONG, FULL, or BSDu. "
            "NORMAL is the format used by ps when no other format specifiers are "
            "given.  LONG is the long format from the BSD versions of ps.  FULL is "
            "the full format used by SysV versions of ps.  BSDu is the 'u' format "
            "used by BSD versions of ps.  \"-print NORMAL\" is implied if no "
            "other action command is GIVEN (i.e. the expression given on the "
            "command line is -and'ed with \"-print NORMAL\", with print appearing "
            "on the right hand side).",
        1,
        OPT_ACTION,
        opt_parse_print,
        opt_eval_print
    },
    {
        "-kill", 
        " <signal> - Sends the given signal to the process and if it succeeds, "
            "evaluates to true, otherwise evaluates to false.",
        1,
        OPT_ACTION,
        opt_parse_kill,
        opt_eval_kill
    },
    {
        "-exec",
        " <command> [arguments] ; - Executes the given command "
            "and returns true if command has a zero exit status and false "
            "otherwise. If \"{}\" appears in the argument list, it is replaced "
            "by the process id of the process.  This occurs in all such "
            "instances.  Note that it may be necessary to 'escape' the semicolon "
            "and/or the french braces on some shells.",
        OPT_VARARGS,
        OPT_ACTION,
        opt_parse_exec,
        opt_eval_exec
    },
    {
        "-cmd",
        " <command> - Evaluates to true on processes whose executable name "
            "matches the given command name.",
        1,
        OPT_TEST,
        opt_parse_strlist,
        opt_eval_cmd
    },
    {
        "-ruid",
        " <uid1>[,uid2[,uid3...]] - Evaluates to true when "
            "the real UID (user id) of a process matches the given uids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_ruid
    },
    {
        "-euid",
        " <uid1>[,uid2[,uid3...]] - Evaluates to true when "
            "the effective UID (user id) of a process matches the given uids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_euid
    },
    {
        "-suid",
        " <uid1>[,uid2[,uid3...]] - Evaluates to true when "
            "the saved UID (user id) of a process matches the given uids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_suid
    },
    {
        "-rgid",
        " <gid1>[,gid2[,gid3...]] - Evaluates to true when "
            "the real GID (group id) of a process matches the given gids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_rgid
    },
    {
        "-egid",
        " <gid1>[,gid2[,gid3...]] - Evaluates to true when "
            "the effective GID (group id) of a process matches the given gids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_egid
    },
    {
        "-sgid",
        " <gid1>[,gid2[,gid3...]] - Evaluates to true when "
            "the saved GID (group id) of a process matches the given gids.",
        1,
        OPT_TEST,
        opt_parse_intlist,
        opt_eval_sgid
    },
    {
        "-nice",
        " <nicelvl> - Evaluates to true when the nice value of a process is "
            "greater than or equal to niceval (numerically).",
        1,
        OPT_TEST,
        opt_parse_int,
        opt_eval_nice
    },
    {
        "-pri",
        " <pri> - Evaluates to true when the priority of a process is "
            "greater than or equal to pri (numerically).",
        1,
        OPT_TEST,
        opt_parse_int,
        opt_eval_pri
    },
    {
        "-runnable",
        " - Evaluates to true when a process is runnable.",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_runnable
    },
    {
        "-sleeping",
        " - Evaluates to true when a process is sleeping (swapped on AIX).",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_sleeping
    },
    {
        "-stopped",
        " - Evaluates to true when a process is stopped or traced.",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_stopped
    },
    {
        "-traced",
        " - Evaluates to true when a process is stopped or traced.",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_stopped
    },
    {
        "-zombie",
        " - Evaluates to true when a process is a zombie (it has exited, but "
            "its parent has not called wait).",
        0,
        OPT_TEST,
        opt_parse_null,
        opt_eval_zombie
    },
};

int num_options;

void opt_init()
{
    num_options = sizeof(options)/sizeof(options[0]);
    qsort(options, num_options, sizeof(option_t), optcmp);
}

#ifdef _GEN_DOCUMENTATION

static int opttypecmp(const void *a, const void *b)
{
    const option_t *x = a, *y = b;

    if (x->opttype < y->opttype)
        return -1;
    if (x->opttype > y->opttype)
        return 1;

    return 0;
}

void opt_print_doctext()
{
    int i;
    num_options = sizeof(options)/sizeof(options[0]);

    /* Sort by option type... */
    qsort(options, num_options, sizeof(option_t), opttypecmp);

    for (i = 0; i < num_options; i++) {
        if (options[i].opttype == OPT_ACTION)
            break;
    }

    /* Then alphabetically */
    qsort(options, i, sizeof(option_t), optcmp);
    qsort(options + i, num_options - i, sizeof(option_t), optcmp);

    for (i = 0; i < num_options; i++) {
        if (i != num_options - 1) 
            printf(".\t");
        else
            printf("-\t");
        printf("%s%s\n", options[i].name, options[i].desc);
    }
}

int main()
{
    opt_print_doctext();

    return 0;
}

#endif /* _GEN_DOCUMENTATION */
