%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <signal.h>
#include "ptfps.h"
#include "get_date.h"
#include "list.h"
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif /* HAVE_CONFIG_H */

#define YYDEBUG 1
#define YYERROR_VERBOSE 1
	
int yylex();

static int argc;
static char **argv;

static int needs_print = 0;
static struct expression *e;

static int yyerror(char *s);

char errormsg[1024];

%}

%union {
	struct expression *exp;
}

%type <exp> args
%type <exp> expression

%token <exp> TERM
%token <exp> ERROR

%left OR
%left AND
%left NOT

%%

args:		/* No args given means true */
			{
				e = $$ = NULL;
			}
			| expression
			{
				e = $$ = $1;
			}
			;

expression:	expression AND expression
			{
				e = $$ = expression_conj(CONJ_AND, $1, $3); 
			}
			| expression OR expression
			{
				e = $$ = expression_conj(CONJ_OR, $1, $3); 
			}
			| NOT expression %prec NOT
			{
				e = $$ = expression_negate($2);
			}
			| '(' expression ')'
			{
				e = $$ = $2;
			}
			| TERM
			{
				e = $$ = $1;
			}
			| ERROR
			{
				expression_destroy(e);
				e = NULL;
				YYABORT;
			}
			;
%%

static int yyerror(char *s)
{
	fprintf(stderr, "%s\n", s);

	return 0;
}


static struct operator {
	char *name;
	int token_type;
}  uniops[] = {
	{"-not", NOT},
	{"!", NOT}
}, binops[] = {
	{"-and", AND},
	{"-a", AND},
	{"-or", OR},
	{"-o", OR},
};

static int num_uniops = sizeof(uniops)/sizeof(uniops[0]);
static int num_binops = sizeof(binops)/sizeof(binops[0]);

int yylex()
{
	int i, j;
	static int c = 2;
	static int implicit_AND_possible = 0;

	/* That's all of the arguments */
	if (c >= argc)
		return 0;

	/* This is an option */
	if (argv[c][0] == '-' || argv[c][0] == '!'){
		/* play some tricks to allow implicit AND's */
		for (i = 0; i < num_binops; i++) {
			if (strcasecmp(argv[c], binops[i].name) == 0) {
				implicit_AND_possible = 0;
				c++;
				return binops[i].token_type;
			}
		}

		if (implicit_AND_possible) {
			implicit_AND_possible = 0;
			return AND;
		}

		for (i = 0; i < num_uniops; i++) {
			if (strcasecmp(argv[c], uniops[i].name) == 0) {
				implicit_AND_possible = 0;
				c++;
				return uniops[i].token_type;
			}
		}

		implicit_AND_possible = 1;

		/* option's */
		for (i = 0; i < num_options; i++){
			if (strcasecmp(argv[c], options[i].name) == 0){
				if (options[i].opttype == OPT_ACTION)
					needs_print = 0;
				c++;
				if (options[i].numargs == OPT_VARARGS) {
					for (j = c; j < argc; j++) {
						if (strcmp(argv[j], OPT_VAEND) == 0) {
							yylval.exp = expression_term(options[i].name, argv+c);
							if (yylval.exp == NULL) {
								return 0;
							}
							c = j;
							return TERM;
						}
					}
				} else {
					yylval.exp = expression_term(options[i].name, argv+c);
					if (yylval.exp == NULL) {
						return 0;
					}
					c += options[i].numargs;
					return TERM;
				}
			}
		}

		snprintf(errormsg, 1024, "Unrecognized option '%s'\n", argv[c]);
		return ERROR;
	}

	/* pass parentheses */
	if (argv[c][1] == '\0'){
		switch (argv[c][0]){
			case '(':
				if (implicit_AND_possible){
					implicit_AND_possible = 0;
					return AND;
				}
				implicit_AND_possible = 0;
				c++;
				return '(';
			case ')':
				implicit_AND_possible = 1;
				c++;
				return ')';
		}
	}

	implicit_AND_possible = 1;

	return 0;
}

struct expression *expression_parse(int ac, char *av[])
{
	argc = ac;
	argv = av;

	if (getenv("DEBUG"))
		yydebug = 1;

	needs_print = 1;

	if (yyparse() != 0){
		return NULL;
	}

	if (needs_print){
		char *normal = "NORMAL";
		struct expression *print, *tmp;

		print = expression_term("-print", &normal);
		if (print == NULL){
			expression_destroy(e);
			return NULL;
		}

		tmp = expression_conj(CONJ_AND, e, print);
		if (tmp == NULL){
			expression_destroy(print);
			expression_destroy(e);
			return NULL;
		}

		e = tmp;
	}

	return e;
}
